{***********************************************************************

Copyright (C) 2016 Bergen A van <x1@ziggo.nl>

This file is part of P3.

P3 is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your
option) any later version.

P3 is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see <http://www.gnu.org/licenses/>.

***********************************************************************}

UNIT p3m6; { sequential file (p3m5) sort }

{ Balanced 2-way merge sort. It uses a variant of heapsort in the
  initial phase for distribution of the unsorted file records to two
  output files.

  The number of disk drives critically affects the efficiency of the
  sort. The ideal situation is to have one auxiliary file associated
  with each disk drive. }

{$I+} {$R+}

INTERFACE

USES p3m3, { element = p3m3_Ttransaction        }
     p3m5; { sequential file of elements (p3m3) }

PROCEDURE p3m6_Rfilesort(diskdrive1,diskdrive2,diskdrive3,diskdrive4:p3m5_Tpathname;
                         unsortedfile,sortedfile                    :p3m5_Pfile);

{**********************************************************************}

IMPLEMENTATION

PROCEDURE p3m6_Rfilesort(diskdrive1,diskdrive2,diskdrive3,diskdrive4:p3m5_Tpathname;
                         unsortedfile,sortedfile                    :p3m5_Pfile);

  TYPE Tfilenumber=0..4;

  VAR scratchfile       :ARRAY[1..4] OF p3m5_Pfile;
      scratchfilename   :p3m5_Tpathname;
      i,sortedfilenumber:Tfilenumber;

  PROCEDURE Rdistribute; { variant of heapsort }

    CONST Cmaxheapsize=256; { record/heap }

    TYPE Toutputfilenumber=0..2;
         Trecordnumber    =0..Cmaxheapsize;

    VAR heap            :ARRAY[1..Cmaxheapsize] OF p3m3_Telement;
        heapsize        :2..Cmaxheapsize;
        midheap         :1..Cmaxheapsize DIV 2;
        inputfile       :p3m5_Pinputfile;
        outputfile      :ARRAY[1..2] OF p3m5_Poutputfile;
        outputfilenumber:Toutputfilenumber;
        x               :p3m3_Telement;
        a,b,j           :Trecordnumber;

    PROCEDURE Rsift(first,last:Trecordnumber);

      VAR siftrecord:p3m3_Telement;
          v1        :Trecordnumber;
          v2        :2..2*Cmaxheapsize;
          q         :BOOLEAN;

      BEGIN
        siftrecord:=heap[first];
        v1:=first;v2:=2*first;
        q:=(v2>last);
        WHILE NOT q DO
          BEGIN
            IF v2<last THEN IF NOT p3m3_Relementorder(heap[v2],heap[v2+1]) THEN v2:=v2+1;
            IF p3m3_Relementorder(siftrecord,heap[v2])
            THEN q:=TRUE
            ELSE BEGIN
                   heap[v1]:=heap[v2];
                   v1:=v2;v2:=2*v1;
                   q:=(v2>last)
                 END
          END;
        heap[v1]:=siftrecord
      END;

    PROCEDURE Rswitchoutputfile;

      BEGIN
        IF outputfilenumber=1
        THEN outputfilenumber:=2
        ELSE outputfilenumber:=1
      END;

    BEGIN { Rdistribute }
      a:=0;
      inputfile:=NEW(p3m5_Pinputfile,Minitialize(unsortedfile));
      inputfile^.Mread(x);
      WHILE NOT inputfile^.Meof AND (a<Cmaxheapsize) DO
        BEGIN
          heap[a+1]:=x;
          a:=a+1;
          inputfile^.Mread(x)
        END;
      outputfile[1]:=NEW(p3m5_Poutputfile,Minitialize(scratchfile[1],100));
      outputfile[2]:=NEW(p3m5_Poutputfile,Minitialize(scratchfile[2],100));
      outputfilenumber:=1;
      IF a=1 THEN outputfile[outputfilenumber]^.Mwrite(heap[1]);
      IF a>=2 THEN
        BEGIN
          heapsize:=a;
          midheap:=heapsize DIV 2;
          FOR j:=midheap DOWNTO 1 DO Rsift(j,heapsize);
          WHILE NOT inputfile^.Meof DO
            BEGIN
              outputfile[outputfilenumber]^.Mwrite(heap[1]);
              IF p3m3_Relementorder(heap[1],x)
              THEN BEGIN
                     heap[1]:=x;
                     Rsift(1,a)
                   END
              ELSE BEGIN
                     heap[1]:=heap[a];
                     Rsift(1,a-1);
                     heap[a]:=x;
                     IF a<=midheap THEN Rsift(a,heapsize);
                     a:=a-1;
                     IF a=0 THEN
                       BEGIN
                         a:=heapsize;
                         Rswitchoutputfile
                      END
                   END;
              inputfile^.Mread(x)
            END;
          b:=heapsize;
          FOR j:=a DOWNTO 1 DO
            BEGIN
              outputfile[outputfilenumber]^.Mwrite(heap[1]);
              heap[1]:=heap[j];
              Rsift(1,j-1);
              heap[j]:=heap[b];
              IF j<=midheap THEN Rsift(j,b-1);
              b:=b-1
            END;
          Rswitchoutputfile;
          IF b>0 THEN
            FOR j:=b DOWNTO 1 DO
              BEGIN
                outputfile[outputfilenumber]^.Mwrite(heap[1]);
                heap[1]:=heap[j];
                Rsift(1,j)
              END
        END;
      DISPOSE(outputfile[2],Mfinalize);
      DISPOSE(outputfile[1],Mfinalize);
      DISPOSE(inputfile,Mfinalize)
    END;

  PROCEDURE Rperformmerges(VAR sortedfilenumber:Tfilenumber);

    VAR a,b,c       :Tfilenumber;
        sortcomplete:BOOLEAN;

    PROCEDURE Rpass(    a,b         :Tfilenumber;
                    VAR sortcomplete:BOOLEAN);

      TYPE Tinputfilenumber =0..2;
           Toutputfilenumber=0..2;

      VAR inputfile       :ARRAY[1..2] OF p3m5_Pinputfile;
          outputfile      :ARRAY[1..2] OF p3m5_Poutputfile;
          outputfilenumber:Toutputfilenumber;
          x1,x2           :p3m3_Telement;
          endofrun        :BOOLEAN;

      PROCEDURE Rswitchoutputfile;

        BEGIN
          IF outputfilenumber=1
          THEN outputfilenumber:=2
          ELSE outputfilenumber:=1
        END;

      PROCEDURE Rcopy(    inputfilenumber:Tinputfilenumber;
                      VAR y2             :p3m3_Telement);

        VAR y1:p3m3_Telement;

        BEGIN
          outputfile[outputfilenumber]^.Mwrite(y2);y1:=y2;
          inputfile[inputfilenumber]^.Mread(y2);
          IF inputfile[inputfilenumber]^.Meof
          THEN endofrun:=TRUE
          ELSE endofrun:=NOT p3m3_Relementorder(y1,y2)
        END;

      PROCEDURE Rcopyrun(    inputfilenumber:Tinputfilenumber;
                         VAR y              :p3m3_Telement);

        BEGIN
          REPEAT Rcopy(inputfilenumber,y) UNTIL endofrun
        END;

      PROCEDURE Rmerge;

        BEGIN
          REPEAT
            IF p3m3_Relementorder(x1,x2)
            THEN BEGIN
                   Rcopy(1,x1);IF endofrun THEN Rcopyrun(2,x2)
                 END
            ELSE BEGIN
                   Rcopy(2,x2);IF endofrun THEN Rcopyrun(1,x1)
                 END
          UNTIL endofrun
        END;

      BEGIN { Rpass }
        inputfile[1]:=NEW(p3m5_Pinputfile,Minitialize(scratchfile[a]));
        inputfile[2]:=NEW(p3m5_Pinputfile,Minitialize(scratchfile[a+1]));
        inputfile[1]^.Mread(x1);
        inputfile[2]^.Mread(x2);
        sortcomplete:=(inputfile[1]^.Meof) OR (inputfile[2]^.Meof);
        IF NOT sortcomplete THEN
          BEGIN
            outputfile[1]:=NEW(p3m5_Poutputfile,Minitialize(scratchfile[b],100));
            outputfile[2]:=NEW(p3m5_Poutputfile,Minitialize(scratchfile[b+1],100));
            outputfilenumber:=1;
            REPEAT
              Rmerge;
              Rswitchoutputfile;
            UNTIL (inputfile[1]^.Meof) OR (inputfile[2]^.Meof);
            WHILE NOT inputfile[1]^.Meof DO
              BEGIN
                Rcopyrun(1,x1);
                Rswitchoutputfile;
              END;
            WHILE NOT inputfile[2]^.Meof DO
              BEGIN
                Rcopyrun(2,x2);
                Rswitchoutputfile;
              END;
            DISPOSE(outputfile[2],Mfinalize);
            DISPOSE(outputfile[1],Mfinalize)
          END;
        DISPOSE(inputfile[2],Mfinalize);
        DISPOSE(inputfile[1],Mfinalize)
      END;

    BEGIN { Rperformmerges }
      sortcomplete:=FALSE;
      a:=1;b:=3;
      REPEAT
        Rpass(a,b,sortcomplete);
        IF NOT sortcomplete THEN
          BEGIN
            c:=a;a:=b;b:=c
          END;
      UNTIL sortcomplete;
      sortedfilenumber:=a
    END;

  PROCEDURE Rcopy(sortedfilenumber:Tfilenumber);

    VAR inputfile :p3m5_Pinputfile;
        outputfile:p3m5_Poutputfile;
        x         :p3m3_Telement;

    BEGIN
      inputfile:=NEW(p3m5_Pinputfile,Minitialize(scratchfile[sortedfilenumber]));
      outputfile:=NEW(p3m5_Poutputfile,Minitialize(sortedfile,100));
      inputfile^.Mread(x);
      WHILE NOT inputfile^.Meof DO
        BEGIN
          outputfile^.Mwrite(x);
          inputfile^.Mread(x)
        END;
      DISPOSE(outputfile,Mfinalize);
      DISPOSE(inputfile,Mfinalize)
    END;

  PROCEDURE Rerasefile(VAR filename:p3m5_Tpathname);

    VAR f:TEXT;

    BEGIN
      ASSIGN(f,filename);
      ERASE(f)
    END;

  BEGIN { p3m6_Rfilesort }
    scratchfile[1]:=NEW(p3m5_Pfile,Minitialize(diskdrive1+'scratch1.bin',TRUE));
    scratchfile[2]:=NEW(p3m5_Pfile,Minitialize(diskdrive2+'scratch2.bin',TRUE));
    scratchfile[3]:=NEW(p3m5_Pfile,Minitialize(diskdrive3+'scratch3.bin',TRUE));
    scratchfile[4]:=NEW(p3m5_Pfile,Minitialize(diskdrive4+'scratch4.bin',TRUE));
    Rdistribute;
    Rperformmerges(sortedfilenumber);
    Rcopy(sortedfilenumber);
    FOR i:=4 DOWNTO 1 DO
      BEGIN
        scratchfilename:=scratchfile[i]^.Pfilename;
        DISPOSE(scratchfile[i],Mfinalize);
        Rerasefile(scratchfilename)
      END
  END;

END.
